#ifndef __JABY_ENGINE_PRIMITIVE_LINE_TYPES_HPP__
#define __JABY_ENGINE_PRIMITIVE_LINE_TYPES_HPP__
#include "primitive_support_types.hpp"

namespace JabyEngine {
    namespace GPU {
        namespace internal {
            struct LineCode : public CodeBase<LineCode> {
                static constexpr uint8_t CmdValue = 0b010;

                static constexpr auto GouraudShading  = Bit(28 - BitCorrection);
                static constexpr auto FlatShading     = !GouraudShading;
                static constexpr auto PolyLine        = Bit(27 - BitCorrection);
                static constexpr auto SingleLine      = !PolyLine;
                static constexpr auto SemiTransparent = Bit(25 - BitCorrection);
                static constexpr auto NonTransparent  = !SemiTransparent;
            };

            template<typename T>
            struct LineCodeInterface {
                typedef ::JabyEngine::GPU::internal::LineCode Code;

                constexpr T& set_poly_line(bool set = true) {
                    if(set) {
                        static_cast<T*>(this)->code.set(Code::PolyLine);
                    }
                    else {
                        static_cast<T*>(this)->code.set(Code::SingleLine);
                    }
                    return *static_cast<T*>(this);
                }

                constexpr T& set_semi_transparent(bool set = true) {
                    if(set) {
                        static_cast<T*>(this)->code.set(Code::SemiTransparent);
                    }
                    else {
                        static_cast<T*>(this)->code.set(Code::NonTransparent);
                    }
                    return *static_cast<T*>(this);
                }
            };

            struct LineHead : public LineCodeInterface<LineHead> {            
                Color24 color;
                Code    code;
            };

            union LineUnion {
                LineHead head;
                Vertex   vertex;
                Color24  color;

                static constexpr LineUnion create_head(const Color24& color, const LineCode& code, bool is_poly) {
                    return LineUnion{.head = LineHead{.color = color, .code = code}.set_poly_line(is_poly)};
                }

                static constexpr LineUnion create_vertex(const Vertex& vertex) {
                    return LineUnion{.vertex = {vertex.x, vertex.y}};
                }

                static constexpr LineUnion create_terminate() {
                    return LineUnion{.vertex = {static_cast<short>(0x5000), static_cast<short>(0x5000)}};
                }
            };

            template<size_t N> 
            struct Line {
                LineUnion data[N];
            };
        }

        struct LINE_F {
            typedef ::JabyEngine::GPU::internal::LineCode  Code;
            typedef ::JabyEngine::GPU::internal::LineUnion LineUnion;

            static constexpr auto IdentityCode = Code().set(Code::FlatShading).set(Code::SingleLine).set(Code::NonTransparent); 

            static constexpr internal::Line<3> create(const Color24& color, const Vertex (&vertices)[2]) {
                return internal::Line<3>{LineUnion::create_head(color, IdentityCode, false), LineUnion::create_vertex(vertices[0]), LineUnion::create_vertex(vertices[1])};
            }

            template<typename...ARGS>
            static constexpr internal::Line<sizeof...(ARGS) + 2> create(const Color24& color, const ARGS&...args) {
                return internal::Line<sizeof...(ARGS) + 2>{LineUnion::create_head(color, IdentityCode, true), LineUnion::create_vertex(args)..., LineUnion::create_terminate()};
            }
        };

        namespace internal {
            template<size_t N>
            struct is_render_primitive<Line<N>> {
                static constexpr bool value = true;
            };
        }
    }
}

#endif //!__JABY_ENGINE_PRIMITIVE_LINE_TYPES_HPP__